// scanner.c 
// Lexical analyer and symbol table for BASICO
//
// The symbol tables are 1-dim char arrays. The symbols
// are space terminated.
// symbol tables are
// keywords for reserved words
// globals for global vars
// locals for for formal parameters and local vars
//
// Search first keywords, then locals, then globals
//
// by Andre Adrian
//
// version 20jun2006
//
// For Modultest:
//  cc -O1 -DMODULTEST -o cscanner scanner.c
 
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "scanner.h"

// scanner
int  Look;          // Lookahead Character
char inputl[128];   // Input line
int  inputi;
char yytext[128];   // identifier or string or number, ..
char token[128];    // identifier or string
int number;         // int const
int  nows;          // No White Space flag

// symtab
char keywords[100];
char tokens[19];
char globals[1000];
char globaltype[100];
char locals[1000];
char localtype[100];
int  localoffset[100];
 
void error(char msg[]) 
{
  inputl[inputi] = 0;
  printf("\n# %s\n", inputl);  
  printf("# error %s\n", msg);
  exit(1);
}
  
// Read New Character From Input Stream (see Crenshaw)
void GetChar() 
{
  if (Look == '\n') {
    inputl[inputi] = 0;
    printf("# %s", inputl);
    inputi = 0;
  }  
  Look = getchar();
  if (Look >= 0) {
    inputl[inputi] = Look;
    inputi = inputi + 1;
  }
}

// Initialize (see Crenshaw)
void scannerInit() 
{
  inputi = 0;
  GetChar();
}

// return 1 if ch is in s, 0 else
int in(char ch, char s[]) 
{
  int i;  
  i = 0;
  while (s[i] != 0) {
    if (ch == s[i]) {
      return 1;
    }
    i = i + 1;
  }
  return 0;
}

// scanner function
int yylex() 
{
  int i;
  char ch;  
  yytext[0] = 0;
  
  // linefeed to ; translation is part of skip white space
  
  // ignore white space
  while (Look == ' ' || Look == '\t' || Look == '\n') {
    if (Look == '\n' && nows) {
      nows = 0;
      return ';';
    }
    GetChar();
  }
  
  // ignore comment
  while (Look == '/') {
    GetChar();
    if (Look != '/') {
      return '/';
    }    
    while (Look != '\n') {
      GetChar();      
    }
    if (nows) {
      nows = 0;
      return ';';
    }
    GetChar();      
    while (Look == ' ' || Look == '\t' || Look == '\n') {
      GetChar();
    }
  }
    
  // one character tokens
  if (in(Look, "-+*/%&|^~,:;()=#<>[]")) {
    nows = 1;
    ch = Look;
    GetChar();
    return ch;
  }
  
  // char const
  if (Look == '\'') {
    nows = 1;
    GetChar();
    if (Look == '\\') {
      GetChar();
      if (Look == 'n') {
        yytext[0] = 10;
      } else if (Look == 't') {
        yytext[0] = 9;
      } else if (Look == '\'') {
        yytext[0] = 39;
      } else if (Look == '\\') {
        yytext[0] = 92;
      } else {
        error("unknown \\ char");
      }
      GetChar();
      if (Look == '\'') {
        GetChar();
        token[0] = yytext[0];
        return 'c';
      } else {
        error("missing \'");
      }        
    } else {
      yytext[0] = Look;
      GetChar();
      if (Look == '\'') {
        GetChar();
        token[0] = yytext[0];
        return 'c';
      } else {
        error("missing \'");
      }        
    } 
  }   

  // array char const
  if (Look == '"') {
    nows = 1; i = 0;
    yytext[i] = Look;
    i = i + 1;
    GetChar();
    while (Look != '"') {
      yytext[i] = Look;
      i = i + 1;
      GetChar();
      if (i > 120) {
        error("string to long");
      }
    }
    yytext[i] = Look;
    i = i + 1;
    GetChar();
    yytext[i] = 0;
    strcpy(token, yytext);
    return 's';
  }
  
  // int const
  if (isdigit(Look)) {
    nows = 1; i = 0;
    yytext[i] = Look;
    i = i + 1;
    GetChar();
    while (isdigit(Look)) {
      yytext[i] = Look;
      i = i + 1;
      GetChar();      
      if (i > 120) {
        error("number too long");
      }
    }
    yytext[i] = 0;
    number = atoi(yytext);
    return 'n';
  }
    
  // keyword or ident 
  if (isalpha(Look)) {
    nows = 1; i = 0;
    yytext[i] = Look;
    i = i + 1;
    GetChar();
    while (isalnum(Look)) {
      yytext[i] = Look;
      i = i + 1;
      GetChar();      
      if (i > 120) {
        error("ident too long");
      }
    }
    yytext[i] = 0;
    ch = gettoken(yytext);
    if (ch == 'x') {
      strcpy(token, yytext);
    }      
    return ch;  
  }
  
  if (Look < 0) {
    return Look;   // EOF
  }
  
  error("Unknown character");     
}

// symbol table functions

// look up name in symbol table
// If found return symbol id else -1
int lookup(char symtab[], char name[]) 
{
  char cn, cs;
  int in, is;
  int found, sym;
  in = 0;  cn = name[in];
  is = 0;  cs = symtab[is];
  found = 1; sym = 0;
  while (cs != 0) {	// table end
    if (cn != cs) {
      found = 0;
    }
    is = is + 1; cs = symtab[is];
    if (cn != 0) {
      in = in + 1; cn = name[in];
    }
    if (cs == ' ') {                    // symbol end
      if (found == 1 && cn == 0) {
      	return sym;
      } else {
	in = 0;  cn = name[in];
	is = is + 1; cs = symtab[is];
	sym = sym + 1;
	found = 1;
      }
    }
  }
  return -1;
}

char gettoken(char name[]) 
{
  int id;
  id = lookup(keywords, name);
  if (id >= 0) {
    return tokens[id];
  } else {
    return 'x';
  }
}

// offset is output variable
int gettype(int offset[], char name[]) 
{
  int id;
  id = lookup(locals, name);
  if (id >= 0) {
    offset[0] = localoffset[id];
    return localtype[id];
  }
  id = lookup(globals, name);
  if (-1 == id) {
    error("variable not found");
  }
  return globaltype[id];
}

void addsym(char symtab[], char name[]) 
{
  if (strlen(symtab) + strlen(name) + 2 <= 1000) {
    strcat(symtab, name);
    strcat(symtab, " ");
  } else {
    error("symtab full");
  }
}

int newname(char symtab[], char name[]) 
{
  int ident;  
  ident = lookup(symtab, name);
  if (ident >= 0) {
    return ident;
  } else {
    addsym(symtab, name);
    return lookup(symtab, name);
  }
}

int addglobal(char name[], char type) 
{
  int id;
  id = newname(globals, name);
  globaltype[id] = type;
  return id;
}

int addlocal(char name[], char type, int offset) 
{
  int id;
  id = newname(locals, name);
  localtype[id] = type;
  localoffset[id] = offset;
  return id;
}

void localsInit() 
{
  locals[0] = 0;
}

void symtabInit() 
{
  strcpy(keywords, "array begin break char do end endif forward func if int else return then var void wend while ");
  strcpy(tokens, "abBCdeEfFiIlrtvVwW");
  globals[0] = 0;
  localsInit();
}

void getname(char symtab[], char name[], int ident) 
{  
  int i, j;  
  i = 0;
  while (ident > 0) {
    if (' ' == symtab[i]) {
      ident = ident - 1;
    }
    i = i + 1;
  }
  j = 0;
  while(symtab[i] != ' ') {
    name[j] = symtab[i];
    j = j + 1;
    i = i + 1;
  }
  name[j] = 0;
}

#ifdef MODULTEST
int main(int argc, char *argv[]) 
{
  int ch;  
  ch = 0;
  scannerInit();
  symtabInit();
  while (1) {
    ch = yylex();
    if (ch < 0) break;
    printf("<%c,%s> ", ch, yytext);
  } 
}
#endif
